% AFMMGPU2 Computational experiment for the paper "Adaptive fast
% multipole method on the GPU".

% Timing of individual parts
% This code investigates the time of each part of the fmm algorithm as a
% function of number of particles (N). All timings except the sort timing
% are performed using CPU sorting
if exist('./examples/data/AFMMgpu2.mat','file')
  warning('Existing file AFMMgpu2.mat NOT overwritten.');
end

clear all
rand('state',20120224);

%Parameters
Nmin=1e3;%minimum
Nmax=5e6;%works up to around 1e7
Nstep=60;
tol=1e-6;
ndirect=45;
Nsets=5;
Naverages=10;
vdummy=rand(1,1e3)+1i*rand(1,1e3);
gdummy=rand(1,1e3);
nn = round(logspace(log10(Nmin),log10(Nmax),Nstep));

for k=1:length(nn)
  tic
  for m=1:Nsets
    vortices=rand(1,nn(k))+1i*rand(1,nn(k));
    gammas=rand(1,nn(k));
    %cuda with CPU sort
    fmm2dCUDAnosort( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');%dummy
    for n=1:Naverages
      cuda_nosort_timing(k,m,n,:)=fmm2dCUDAnosort( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
    %full GPU code
    fmm2dCUDA( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');%dummy
    for n=1:Naverages
      cuda_timing(k,m,n,:)=fmm2dCUDA( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
    %cpu code without multipole eval to determine direct eval timing
    fmm2dCPUnompeval( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    for n=1:Naverages
      cpu_nompeval_timing(k,m,n,:)=fmm2dCPUnompeval( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
    %cpu without direct eval to determine multipole eval timing
    fmm2dCPUnodirect( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    for n=1:Naverages
      cpu_nodirect_timing(k,m,n,:)=fmm2dCPUnodirect( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
  end
  disp(['step ' num2str(k) ' points ' num2str(nn(k)) ' time ' num2str(toc)]);
end

if ~exist('./examples/data/AFMMgpu2.mat','file')
  save examples/data/AFMMgpu2 cuda_nosort_timing cuda_timing cpu_nompeval_timing cpu_nodirect_timing nn
end
