function AFMMgpu2plots(all,errorplot)
% AFMMGPU2PLOTS Plots for the paper "Adaptive fast multipole method on
% the GPU".

if nargin<1
  all=false;
end
if nargin<2
  errorplot=false;
end

%Timing of individual parts
plotparams={'LineWidth',0.75};
load examples/data/AFMMgpu2
line1='-k';
line2='-.k';
line3='-.r';
line4='-g';
line5='.-k';
line6='--m';
line7='.-b';
line8='--m';
%calculate averages for each number of points
% cuda_nosort_timing2=mean(cuda_nosort_timing,3);
% cuda_timing2=mean(cuda_timing,3);
% cpu_nompeval_timing2=mean(cpu_nompeval_timing,3);
% cpu_nodirect_timing2=mean(cpu_nodirect_timing,3);
% cuda_nosort_timing3=mean(cuda_nosort_timing2,2);
% cuda_timing3=mean(cuda_timing2,2);
% cpu_nompeval_timing3=mean(cpu_nompeval_timing2,2);
% cpu_nodirect_timing3=mean(cpu_nodirect_timing2,2);
[cuda_nosort_timing3 cuda_nosort_timing3stdev]=AFMMgpu_post_process(cuda_nosort_timing,1,1);
[cuda_timing3 cuda_timing3stdev]=AFMMgpu_post_process(cuda_timing,1,1);
[cpu_nompeval_timing3 cpu_nompeval_timing3stdev]=AFMMgpu_post_process(cpu_nompeval_timing,1,1);
[cpu_nodirect_timing3 cpu_nodirect_timing3stdev]=AFMMgpu_post_process(cpu_nodirect_timing,1,1);

if all
  figure
  loglog(nn,cpu_nompeval_timing3(:,2),line1,nn,cuda_timing3(:,2),line2);
  legend('CPU time','GPU time');
  title('Sorting time');
  xlabel('Number of points');
  ylabel('Time (s)');
  
  figure
  semilogx(nn,cpu_nompeval_timing3(:,2)./cuda_timing3(:,2),line1);
  title('Sorting time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nompeval_timing3(:,3),line1,nn,cuda_nosort_timing3(:,3),line2);
  legend('CPU time','GPU time');
  title('Connectivity time');
  xlabel('Number of points');
  ylabel('Time (s)');
  figure
  semilogx(nn,cpu_nompeval_timing3(:,3)./cuda_nosort_timing3(:,3),line1);
  title('Connectivity time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nompeval_timing3(:,5),line1,nn,cuda_nosort_timing3(:,23),line2);
  legend('CPU time','GPU time');
  title('Initialization time');
  xlabel('Number of points');
  ylabel('Time (s)');
  figure
  semilogx(nn,cpu_nompeval_timing3(:,5)./cuda_nosort_timing3(:,23),line1);
  title('Initialization time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nompeval_timing3(:,6),line1,nn,cuda_nosort_timing3(:,24),line2);
  legend('CPU time','GPU time');
  title('m2m time');
  xlabel('Number of points');
  ylabel('Time (s)');
  
  figure
  semilogx(nn,cpu_nompeval_timing3(:,6)./cuda_nosort_timing3(:,24),line1);
  title('m2m time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nompeval_timing3(:,7),line1,nn,cuda_nosort_timing3(:,25),line2);
  legend('CPU time','GPU time');
  title('m2ps time');
  xlabel('Number of points');
  ylabel('Time (s)');
  
  figure
  semilogx(nn,cpu_nompeval_timing3(:,7)./cuda_nosort_timing3(:,25),line1);
  title('m2ps time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nompeval_timing3(:,8),line1,nn,cuda_nosort_timing3(:,26),line2);
  legend('CPU time','GPU time');
  title('p2p time');
  xlabel('Number of points');
  ylabel('Time (s)');
  
  figure
  semilogx(nn,cpu_nompeval_timing3(:,8)./cuda_nosort_timing3(:,26),line1);
  title('p2p time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nodirect_timing3(:,9),line1,nn,cuda_nosort_timing3(:,27),line2);
  legend('CPU time','GPU time');
  title('multipole evaluation time');
  xlabel('Number of points');
  ylabel('Time (s)');
  
  figure
  semilogx(nn,cpu_nodirect_timing3(:,9)./cuda_nosort_timing3(:,27),line1);
  title('multipole evaluation time');
  xlabel('Number of points');
  ylabel('Speedup');
  
  figure
  loglog(nn,cpu_nompeval_timing3(:,9),line1,nn,cuda_nosort_timing3(:,28),line2);
  legend('CPU time','GPU time');
  title('direct evaluation time');
  xlabel('Number of points');
  ylabel('Time (s)');
  
  figure
  semilogx(nn,cpu_nompeval_timing3(:,9)./cuda_nosort_timing3(:,28),line1);
  title('direct evaluation time');
  xlabel('Number of points');
  ylabel('Speedup');
end
% production quality figure:
figure
semilogx(nn,cpu_nompeval_timing3(:,2)./cuda_timing3(:,2),line1,plotparams{:});
hold on;
semilogx(nn,cpu_nompeval_timing3(:,3)./cuda_nosort_timing3(:,3),line2,plotparams{:});
semilogx(nn,cpu_nodirect_timing3(:,9)./cuda_nosort_timing3(:,27),line7,plotparams{:});
semilogx(nn,cpu_nompeval_timing3(:,9)./cuda_nosort_timing3(:,28),line8,plotparams{:});
if errorplot
  hold on;
  nw = 20; % number of errorbars
  w = floor(numel(nn)/nw);
  eline1 = [line1(end) '.']; % plot only bars and no graphs
  eline2 = [line2(end) '.'];
  eline7 = [line7(end) '.'];
  eline8 = [line8(end) '.'];
  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,2), ...
                               cuda_timing3(:,2), ...
                               cpu_nompeval_timing3stdev(:,2), ...
                               cuda_timing3stdev(:,2));
  h1 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline1);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,3), ...
                               cuda_nosort_timing3(:,3), ...
                               cpu_nompeval_timing3stdev(:,3), ...
                               cuda_nosort_timing3stdev(:,3));
  h2 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline2);
  [timing,stdev] = getdivstdev(cpu_nodirect_timing3(:,9), ...
                               cuda_nosort_timing3(:,27), ...
                               cpu_nodirect_timing3stdev(:,9), ...
                               cuda_nosort_timing3stdev(:,27));
  h3 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline7);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,9), ...
                               cuda_nosort_timing3(:,28), ...
                               cpu_nompeval_timing3stdev(:,9), ...
                               cuda_nosort_timing3stdev(:,28));
  h4 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline8);
  
  errorbar_tick(h1,0.25*nn(1:w:end),'units');
  errorbar_tick(h2,0.25*nn(1:w:end),'units');
  errorbar_tick(h3,0.25*nn(1:w:end),'units');
  errorbar_tick(h4,0.25*nn(1:w:end),'units');
end
h = legend('Sort','Connect','L2P','P2P','Location','NW');
set(h,'FontSize',10);
h = xlabel('N'); set(h,'FontSize',10);
h = ylabel('Speedup'); set(h,'FontSize',10);
axis([1e3 5e6 0 20]);
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 220]);
print -depsc examples/fig/nspeedup1.eps

figure
semilogx(nn,cpu_nompeval_timing3(:,5)./cuda_nosort_timing3(:,23),line3,plotparams{:});
hold on;
semilogx(nn,cpu_nompeval_timing3(:,6)./cuda_nosort_timing3(:,24),line4,plotparams{:});
semilogx(nn,cpu_nompeval_timing3(:,7)./cuda_nosort_timing3(:,25),line5,plotparams{:});
semilogx(nn,cpu_nompeval_timing3(:,8)./cuda_nosort_timing3(:,26),line6,plotparams{:});
if errorplot
  hold on;
  nw = 20; % number of errorbars
  w = floor(numel(nn)/nw);
  eline3 = [line3(end) '.']; % plot only bars and no graphs
  eline4 = [line4(end) '.'];
  eline5 = [line5(end) '.'];
  eline6 = [line6(end) '.'];

  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,5), ...
                               cuda_nosort_timing3(:,23), ...
                               cpu_nompeval_timing3stdev(:,5), ...
                               cuda_nosort_timing3stdev(:,23));
  h1 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline3);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,6), ...
                               cuda_nosort_timing3(:,24), ...
                               cpu_nompeval_timing3stdev(:,6), ...
                               cuda_nosort_timing3stdev(:,24));
  h2 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline4);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,7), ...
                               cuda_nosort_timing3(:,25), ...
                               cpu_nompeval_timing3stdev(:,7), ...
                               cuda_nosort_timing3stdev(:,25));
  h3 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline5);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing3(:,8), ...
                               cuda_nosort_timing3(:,26), ...
                               cpu_nompeval_timing3stdev(:,8), ...
                               cuda_nosort_timing3stdev(:,26));
  h4 = errorbar(nn(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline6);

  errorbar_tick(h1,0.25*nn(1:w:end),'units');
  errorbar_tick(h2,0.25*nn(1:w:end),'units');
  errorbar_tick(h3,0.25*nn(1:w:end),'units');
  errorbar_tick(h4,0.25*nn(1:w:end),'units');
end
h = legend('P2M','M2M','M2L','L2L','Location','NW');
set(h,'FontSize',10);
h = xlabel('N'); set(h,'FontSize',10);
h = ylabel('Speedup'); set(h,'FontSize',10);
axis([1e3 5e6 0 20]);
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 220]);
print -depsc examples/fig/nspeedup2.eps
