% AFMMGPU3 Computational experiment for the paper "Adaptive fast
% multipole method on the GPU".

% Timing as a function of ndirect
% This code fixes the number of multipole levels to 8 and uses different
% number of particles to obtain different numbers of elements in each box.
if exist('./examples/data/AFMMgpu3.mat','file')
  warning('Existing file AFMMgpu3.mat NOT overwritten.');
end

clear all
rand('state',20120225);

%parameters
ndirmin=1; %minimum ndirect
ndirmax=80;%maximum ndirect
tol=1e-6;
Nsets=5;
Naverages=10;
nn = round((ndirmin:ndirmax)*2^16);

for k=1:length(nn)
  tic
  ndirect=round(nn(k)/2^15*5/8); %should give 8 levels
  for m=1:Nsets
    vortices=rand(1,nn(k))+1i*rand(1,nn(k));
    gammas=rand(1,nn(k));
    %cuda with CPU sort
    fmm2dCUDAnosort( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cuda_nosort_timing_deval(k,m,n,:)=fmm2dCUDAnosort( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
    %full GPU code
    fmm2dCUDA( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cuda_timing_deval(k,m,n,:)=fmm2dCUDA( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
    %cpu code without multipole eval to determine direct eval timing
    fmm2dCPUnompeval( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cpu_nompeval_timing_deval(k,m,n,:)=fmm2dCPUnompeval( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
    %cpu without direct eval to determine multipole eval timing
    fmm2dCPUnodirect( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cpu_nodirect_timing_deval(k,m,n,:)=fmm2dCPUnodirect( vortices,gammas,'pot',1,'ndirect',ndirect,'tol',tol,'out','time');
    end
  end
  disp(['step ' num2str(k) ' points ' num2str(nn(k)) ' time ' num2str(toc)]);
end

if ~exist('./examples/data/AFMMgpu3.mat','file')
  save examples/data/AFMMgpu3 cuda_timing_deval cuda_nosort_timing_deval cpu_nompeval_timing_deval cpu_nodirect_timing_deval nn ndirmin ndirmax
end
