% AFMMGPU4 Computational experiment for the paper "Adaptive fast
% multipole method on the GPU".

% Find ndirect as function of tol
% This code performes AFMMgpu0 for several tolerances and calculates
% optimal ndirect for each of them. 
if exist('./examples/data/AFMMgpu4.mat','file')
  warning('Existing file AFMMgpu4.mat NOT overwritten.');
end

clear all
rand('state',20120226);

%Parameters
Nmin=2e4;%minimum
Nmax=5e5;
pmin=2;
pmax=52;
theta=0.5;
Nstep=50;
pp=pmin:pmax;
tols=2.^((pp+0.5).*log2(theta)-2);
Nsets=5;
Naverages=5;
nn = round(logspace(log10(Nmin),log10(Nmax),Nstep));
ndir=10:75;

for k=1:length(nn)
  for m=1:Nsets
    vortices{k,m}=rand(1,nn(k))+1i*rand(1,nn(k));
    gammas{k,m}=rand(1,nn(k));
  end
end

for u=1:length(tols)
  disp(['tol = ' num2str(tols(u))]);
  for k=1:length(nn)
    tic
    level=0;
    for h=1:length(ndir)
      %check if new value would give same number of levels. In that case,
      %copy results from last run
      newlevel=ceil(0.5*log2(5/8*nn(k)/ndir(h)));
      if newlevel==level && h ~=1
        cuda_ndirect_tol(u,h,k,:,:,:)=cuda_ndirect_tol(u,h-1,k,:,:,:);
        cpu_ndirect_tol(u,h,k,:,:,:)=cpu_ndirect_tol(u,h-1,k,:,:,:);
      else
        level=newlevel;
        for m=1:Nsets
          fmm2dCUDA( vortices{k,m},gammas{k,m},'pot',1,'ndirect',ndir(h),'tol',tols(u),'out','time');
          for n=1:Naverages
            cuda_ndirect_tol(u,h,k,m,n,:)=fmm2dCUDA( vortices{k,m},gammas{k,m},'pot',1,'ndirect',ndir(h),'tol',tols(u),'out','time');
          end
          fmm2dCPU( vortices{k,m},gammas{k,m},'pot',1,'ndirect',ndir(h),'tol',tols(u),'out','time');
          for n=1:Naverages
            cpu_ndirect_tol(u,h,k,m,n,:)=fmm2dCPU( vortices{k,m},gammas{k,m},'pot',1,'ndirect',ndir(h),'tol',tols(u),'out','time');
          end
        end
      end
    end
    disp(['step ' num2str(k) ' points ' num2str(nn(k)) ' time ' num2str(toc)]);
  end
end

if ~exist('./examples/data/AFMMgpu4.mat','file')
  save examples/data/AFMMgpu4 cuda_ndirect_tol cpu_ndirect_tol nn ndir tols pp
end
