function AFMMgpu5plots(all,errorplot)
% AFMMGPU5PLOTS Plots for the paper "Adaptive fast multipole method on
% the GPU".

if nargin<1
  all=false;
end
if nargin<2
  errorplot=false;
end
%Timing of individual parts
load examples/data/AFMMgpu5
plotparams={'LineWidth',0.75};
line1='-k';
line2='--k';
line3='-.r';
line4='-g';
line5='.-k';
line6='--m';
line7='.-b';
line8='--m';

%calculate averages for each number of points
% cuda_nosort_timing_tol2=mean(cuda_nosort_timing_tol,3);
% cuda_timing_tol2=mean(cuda_timing_tol,3);
% cpu_nompeval_timing_tol2=mean(cpu_nompeval_timing_tol,3);
% cpu_nodirect_timing_tol2=mean(cpu_nodirect_timing_tol,3);
% cuda_nosort_timing_tol3=mean(cuda_nosort_timing_tol2,2);
% cuda_timing_tol3=mean(cuda_timing_tol2,2);
% cpu_nompeval_timing_tol3=mean(cpu_nompeval_timing_tol2,2);
% cpu_nodirect_timing_tol3=mean(cpu_nodirect_timing_tol2,2);
[cuda_nosort_timing_tol3 cuda_nosort_timing_tol3stdev]=AFMMgpu_post_process(cuda_nosort_timing_tol,1,1);
[cuda_timing_tol3 cuda_timing_tol3stdev]=AFMMgpu_post_process(cuda_timing_tol,1,1);
[cpu_nompeval_timing_tol3 cpu_nompeval_timing_tol3stdev]=AFMMgpu_post_process(cpu_nompeval_timing_tol,1,1);
[cpu_nodirect_timing_tol3 cpu_nodirect_timing_tol3stdev]=AFMMgpu_post_process(cpu_nodirect_timing_tol,1,1);

if all
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,2),line1,pp,cuda_timing_tol3(:,2),line2);
  legend('CPU time','GPU time');
  title('Sorting time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,2)./cuda_timing_tol3(:,2),line1);
  title('Sorting time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,3),line1,pp,cuda_nosort_timing_tol3(:,3),line2);
  legend('CPU time','GPU time');
  title('Connectivity time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,3)./cuda_nosort_timing_tol3(:,3),line1);
  title('Connectivity time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,5),line1,pp,cuda_nosort_timing_tol3(:,23),line2);
  legend('CPU time','GPU time');
  title('Initialization time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,5)./cuda_nosort_timing_tol3(:,23),line1);
  title('Initialization time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,6),line1,pp,cuda_nosort_timing_tol3(:,24),line2);
  legend('CPU time','GPU time');
  title('m2m time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,6)./cuda_nosort_timing_tol3(:,24),line1);
  title('m2m time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,7),line1,pp,cuda_nosort_timing_tol3(:,25),line2);
  legend('CPU time','GPU time');
  title('m2ps time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,7)./cuda_nosort_timing_tol3(:,25),line1);
  title('m2ps time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,8),line1,pp,cuda_nosort_timing_tol3(:,26),line2);
  legend('CPU time','GPU time');
  title('p2p time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,8)./cuda_nosort_timing_tol3(:,26),line1);
  title('p2p time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nodirect_timing_tol3(:,9),line1,pp,cuda_nosort_timing_tol3(:,27),line2);
  legend('CPU time','GPU time');
  title('multipole evaluation time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nodirect_timing_tol3(:,9)./cuda_nosort_timing_tol3(:,27),line1);
  title('multipole evaluation time');
  xlabel('Tolerance');
  ylabel('Speedup');
  
  figure
  semilogy(pp,cpu_nompeval_timing_tol3(:,9),line1,pp,cuda_nosort_timing_tol3(:,28),line2);
  legend('CPU time','GPU time');
  title('direct evaluation time');
  xlabel('Tolerance');
  ylabel('Time (s)');
  figure
  plot(pp,cpu_nompeval_timing_tol3(:,9)./cuda_nosort_timing_tol3(:,28),line1);
  title('direct evaluation time');
  xlabel('Tolerance');
  ylabel('Speedup');
end

% production quality figure:
figure, hold on;
plot(pp,cpu_nompeval_timing_tol3(:,5)./cuda_nosort_timing_tol3(:,23),line3,plotparams{:});
plot(pp,cpu_nompeval_timing_tol3(:,6)./cuda_nosort_timing_tol3(:,24),line4,plotparams{:});
plot(pp,cpu_nompeval_timing_tol3(:,7)./cuda_nosort_timing_tol3(:,25),line5,plotparams{:});
plot(pp,cpu_nompeval_timing_tol3(:,8)./cuda_nosort_timing_tol3(:,26),line6,plotparams{:});
plot(pp,cpu_nodirect_timing_tol3(:,9)./cuda_nosort_timing_tol3(:,27),line7,plotparams{:});
if errorplot
  hold on;
  nw = 20; % number of errorbars
  w = floor(numel(pp)/nw);
  eline3 = [line3(end) '.']; % plot only bars and no graphs
  eline4 = [line4(end) '.'];
  eline5 = [line5(end) '.'];
  eline6 = [line6(end) '.'];
  eline7 = [line7(end) '.'];

  [timing,stdev] = getdivstdev(cpu_nompeval_timing_tol3(:,5), ...
                               cuda_nosort_timing_tol3(:,23), ...
                               cpu_nompeval_timing_tol3stdev(:,5), ...
                               cuda_nosort_timing_tol3stdev(:,23));
  h1 = errorbar(pp(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline3);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing_tol3(:,6), ...
                               cuda_nosort_timing_tol3(:,24), ...
                               cpu_nompeval_timing_tol3stdev(:,6), ...
                               cuda_nosort_timing_tol3stdev(:,24));
  h2 = errorbar(pp(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline4);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing_tol3(:,7), ...
                               cuda_nosort_timing_tol3(:,25), ...
                               cpu_nompeval_timing_tol3stdev(:,7), ...
                               cuda_nosort_timing_tol3stdev(:,25));
  h3 = errorbar(pp(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline5);
  [timing,stdev] = getdivstdev(cpu_nompeval_timing_tol3(:,8), ...
                               cuda_nosort_timing_tol3(:,26), ...
                               cpu_nompeval_timing_tol3stdev(:,8), ...
                               cuda_nosort_timing_tol3stdev(:,26));
  h4 = errorbar(pp(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline6);
  [timing,stdev] = getdivstdev(cpu_nodirect_timing_tol3(:,9), ...
                               cuda_nosort_timing_tol3(:,27), ...
                               cpu_nodirect_timing_tol3stdev(:,9), ...
                               cuda_nosort_timing_tol3stdev(:,27));
  h5 = errorbar(pp(1:w:end),timing(1:w:end),0.5*stdev(1:w:end),eline7);

  errorbar_tick(h1,2,'units');
  errorbar_tick(h2,2,'units');
  errorbar_tick(h3,2,'units');
  errorbar_tick(h4,2,'units');
  errorbar_tick(h5,2,'units');
end
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 250]);
h = legend('P2M','M2M','M2L','L2L','L2P');
set(h,'FontSize',10);
set(h,'Position',[0.7067 0.4542 0.1836 0.385])
h = xlabel('p'); set(h,'FontSize',10);
h = ylabel('Speedup'); set(h,'FontSize',10);
axis([pp([1 end]) 0 27])
set(gca,'xtick',0:10:50,'ytick',[0:5:25]);
set(gca,'Box','on')
print -depsc examples/fig/pspeedup.eps
