% AFMMGPU6PLOTS Plots for the paper "Adaptive fast multipole method on
% the GPU".

% S. Engblom 2012-03-27

load examples/data/AFMMgpu6
plotparams = {'LineWidth',0.75};

figure(1),
h = loglog(nn,t(1,:),'g-',plotparams{:}); hold on
h = loglog(nn,t(2,:),'b--',plotparams{:});
h = loglog(nn,t(3,:),'r-.',plotparams{:});

axis([1e4 2e7 1e-3 1e2])
h = xlabel('N'); set(h,'FontSize',10);
h = ylabel('Time (s)'); set(h,'FontSize',10);

h = legend('Uniform','Normal','Layer','Location','SouthEast');
set(h,'FontSize',10);
hold on, loglog(nn([5 14]),0.025e-4*nn([5 14]),'k','LineWidth',2)
h = text(3e4,0.2,'\propto N');
set(h,'FontSize',10);

set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 200]);

print -depsc examples/fig/adaptivity.eps

figure(2),
semilogx(sigma2,tgpu1./tgpu0,'b',sigma2,tcpu1./tcpu0,'r--', ...
         sigma2,tgpu2./tgpu0,'b',sigma2,tcpu2./tcpu0,'r--',plotparams{:});
h = xlabel('\sigma'); set(h,'FontSize',10);
h = ylabel('Time'); set(h,'FontSize',10);
h = legend('GPU','CPU'); set(h,'FontSize',10);
axis([sigma2([end 1]) 0 6])
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 200]);

print -depsc examples/fig/nonuniform.eps

% illustrative plot of mesh

% arguments
ndirect = 45;

% reproducible results
randn('state',20120327);

% build distributions
N = 2^24;
m = ones(N,1);
x = zeros(1,0);
y = zeros(1,0);
while numel(x) < N || numel(y) < N
  x = [x; abs(1/10*randn(N-numel(x),1))];
  y = [y; abs(1/10*randn(N-numel(y),1))];
  x(x > 1) = [];
  y(y > 1) = [];
end
z = complex(x,y);
[M,C] = fmm2d(z,m,'ndirect',ndirect,'out','mesh');
l = 5;
M = M{l};
C = C{l};

% 3D-view of box area
a = real(M(2,:)).*imag(M(2,:));
figure(3), meshplot3(M,1./a,'k');
axis square, axis([0 0.6 0 0.6]);
view([55.5 50]);
set(gca,'xtick',[],'ytick',[],'ztick',[]);
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 230 230]);
print -deps examples/fig/mesh3D.eps

% flat view with connectivity pattern
figure(4), meshplot(M,'k'); hold on

% box i
i = 112;
MM = M(:,i);
zz = MM([1 1 1 1],:)+[MM(2,:); conj(MM(2,:)); -MM(2,:); -conj(MM(2,:))];
fill(real(zz),imag(zz),'k');

% direct interaction of box i
j = find(C(:,i) == 1);
j(j == i) = [];
MM = M(:,j);
zz = MM([1 1 1 1],:)+[MM(2,:); conj(MM(2,:)); -MM(2,:); -conj(MM(2,:))];
fill(real(zz),imag(zz),[0.5 0.5 0.5]);

% cluster-to-cluster interaction of box i
j = find(C(:,i) == -1);
MM = M(:,j);
zz = MM([1 1 1 1],:)+[MM(2,:); conj(MM(2,:)); -MM(2,:); -conj(MM(2,:))];
fill(real(zz),imag(zz),[0.8 0.8 0.8]);

axis square, axis([0 0.6 0 0.6])
set(gca,'xtick',[],'ytick',[]);
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 230 230]);
print -deps examples/fig/mesh.eps
