%AFMMGPU_COMPILE Compiles several versions of the FMM.

% A. Goude 2012-09-27

debug_defs = {'-DEBUGPANELSORT' ...       % (1)
              '-DNOMULTIPOLEEVAL'...      % (2) %removes multipoleevaluation, for speed measurements
              '-DNODIRECTEVAL'...         % (3) %removes direct evaluation, for speed measurements
             };
platform_defs = {'-DmwIndex=int' ...      % (1)
                 '-DmwSize=int' ...       % (2)
                };
fmm_defs = {'-DINLINECOMPLEX' ...         % (1)
            '-DINLINECOMPLEX_DIRECT' ...  % (2)
            '-DTHETACUTOFFCHECK' ...      % (3)
            '-DFMM_THETA=0.5' ...         % (4)
            '-DFMM_MESHRATIO=2.0' ...     % (5)
           };
panel_defs = {'-DPANELSORT' ...           % (1)
              '-DPANELDUMMYFACTOR=2' ...  % (2)
              '-DPANELSHRINKBOX' ...      % (3)
             };
exp_defs = {'-DRADIALSHRINK' ...          % (1)
            '-DRADIALSHRINKLIMIT=512' ... % (2)
            '-DCUDATIMESORT' ...          % (3)
            '-DASYMMETRIC_CPU' ...        % (4)
            '-DUNROLLHORNER' ...          % (5)
            '-DUNROLLLEVEL=6' ...         % (6)
            '-DSSE2INTRINSIC' ...         % (7)
            '-DSSE2M2PSSCALING' ...       % (8)
            '-DSSE2DIRECT' ...            % (9)
            '-DSSEEVALUNROLL' ...         % (10)
            '-DSSE2INIT' ...              % (11)
            '-DCROSSORDER' ...            % (12)
            '-DFULLCROSSORDER' ...        % (13)
            '-DUNROLLHORNEROLD' ...       % (14)
            '-DM2PSINBOX'};               % (15)
cuda_defs = {'-DCUDATIME' ...           % (1)
             '-DCUDASUPPORT' ...        % (2)
             '-DMULTIPOLEEVALCUDA' ...  % (3)
             '-DMULTIPOLEINITCUDA' ...  % (4)
             '-DMULTIPOLESHIFTCUDA' ... % (5)
             '-DCUDASORT' ...           % (6)
             };
%for normal CUDA simulation
CUDA_args=[debug_defs([]) platform_defs([]) fmm_defs ...
              panel_defs exp_defs(1:2) cuda_defs(1:6)];
%for evaluation of individual parts of the GPU/CPU code, use CPU sort
CUDA_args_nosort=[debug_defs([]) platform_defs([]) fmm_defs ...
              panel_defs exp_defs(1:2) cuda_defs(1:5)];
%for normal CPU simulation
CPU_args=[debug_defs([]) platform_defs([]) fmm_defs ...
              panel_defs exp_defs([1:2 5:7 8 9:11 13 15]) cuda_defs(1)];
%CPU without mpeval, to measure speed of direct eval
CPU_args_nompeval=[debug_defs(2) platform_defs([]) fmm_defs ...
              panel_defs exp_defs([1:2 5:7 8 9:11 13 15]) cuda_defs(1)];
%CPU without direct eval, to measure speed of mpeval
CPU_args_nodirect=[debug_defs(3) platform_defs([]) fmm_defs ...
              panel_defs exp_defs([1:2 5:7 8 9:11 13 15]) cuda_defs(1)];
            
make('gpu',CUDA_args{:});
movefile(['fmm2d.' mexext],['fmm2dCUDA.' mexext]);
make('gpu',CUDA_args_nosort{:});
movefile(['fmm2d.' mexext],['fmm2dCUDAnosort.' mexext]); 
make('gpu',CPU_args{:});
movefile(['fmm2d.' mexext],['fmm2dCPU.' mexext]); 
make('gpu',CPU_args_nompeval{:});
movefile(['fmm2d.' mexext],['fmm2dCPUnompeval.' mexext]); 
make('gpu',CPU_args_nodirect{:});
movefile(['fmm2d.' mexext],['fmm2dCPUnodirect.' mexext]); 