function [output stdev]=AFMMgpu_post_process(input,preind,postind,keep)
% AFMMgpu_post_process, code used to calculate averages for the timing for
% AFMMgpuplotsX. Code will sort the times and keep the amount of points
% corresponding to the keep parameter, i.e. the default 0.8 will remove the
% 10 % highest and the 10 % lowest values.
% Usage:
% input is the data
% preind, the number of indices ahead of the data to be averaged, works
% with 1, 2 and 3
% postind, the number of indices after the data to be averaged. Works with
% 0 and 1.
% avaraging works with 1 and 2 indices, i.e. Nsets and Naverages
if nargin<4
  keep=0.8; %default
end
%1, 2 or 3 indices to loop over
if preind==0
  output=process1(input,postind,keep);
elseif preind==1
  if length(size(input)) == 2
    for k=1:size(input,1)
      [output(k,:) stdev(k,:)]=process1(input(k,:),postind,keep);
    end
  else
    for k=1:size(input,1)
      aa=size(input);
      [output(k,:) stdev(k,:)]=process1(reshape(input(k,:),aa(2:end)),postind,keep);
    end
  end
elseif preind==2
  for k=1:size(input,1)
    for m=1:size(input,2)
      aa=size(input);
      [output(k,m,:) stdev(k,m,:)] =process1(reshape(input(k,m,:),aa(3:end)),postind,keep);
    end
  end
elseif preind==3
  for k=1:size(input,1)
    for m=1:size(input,2)
      for n=1:size(input,3)
        aa=size(input);
        [output(k,m,n,:) stdev(k,m,n,:)]=process1(reshape(input(k,m,n,:),aa(4:end)),postind,keep);
      end
    end
  end
end

%intermediate function, chooses to loop over the last index or not
function [out stdev]=process1(wp,postind,keep)
if postind==0
  [out stdev]=process(wp,keep);
elseif postind==1
  if length(size(wp))==3 %the trick with reshape only works for indices in the beginning
    for m=1:size(wp,3)
      [out(m) stdev(m)]=process(wp(:,:,m),keep);
    end
  elseif length(size(wp))==2
    for m=1:size(wp,2)
      [out(m) stdev(m)]=process(wp(:,m),keep);
    end
  end
end

%the averaging code.
function [wp stdev]=process(wp,keep)
wp=wp(:);
wp=sort(wp);
remove=round(length(wp)*(1-keep)/2);
wp=wp((1+remove):(end-remove));
stdev=std(wp);
wp=mean(wp);
