function h = meshplot3(M,Z,varargin)
%MESHPLOT3 Plot FMM mesh in 3D.
%   MESHPLOT3(M,Z) plots the FMM mesh M as obtained via the 'out'
%   property in FMM2D. For each box i in M, the height value Z(i) is
%   understood such that the resulting plot is three-dimensional.
%
%   H = MESPLOT3(M,Z...) sends additional arguments to PLOT and
%   returns a handle to the resulting plot.
%
%   Example:
%     N = 10000;
%     z = complex(randn(N,1),randn(N,1));
%     m = ones(N,1);
%
%     [M,C] = fmm2d(z,m,'tol',1e-6,'out','mesh');
%     l = 6;
%     a = real(M{l}(2,:)).*imag(M{l}(2,:)); % area of box
%     a = max(a,1e-4); % skip collapsed boxes
%     figure, meshplot3(M{l},1./a,'k');
%     axis([-4 4 -4 4 0 1e3]);
%
%   See also FMM2D, MESHPLOT, PLOT3.

% S. Engblom 2012-05-09

z = M([1 1 1 1],:)+[M(2,:); conj(M(2,:)); -M(2,:); -conj(M(2,:))];
x = real(z);
y = imag(z);
Z = reshape(Z,1,[]);
Z = Z(ones(1,5),:);
if nargout > 0
  h = plot3(x([1:end 1],:),y([1:end 1],:),Z,varargin{:});
else
  plot3(x([1:end 1],:),y([1:end 1],:),Z,varargin{:});
end
